@@ -5,6 +5,7 @@
 
 #include "pgroonga.h"
 
+#include <access/reloptions.h>
 #include <access/relscan.h>
 #include <catalog/catalog.h>
 #include <catalog/index.h>
@@ -13,6 +14,7 @@
 #include <miscadmin.h>
 #include <storage/ipc.h>
 #include <storage/lmgr.h>
+#include <utils/lsyscache.h>
 #include <utils/selfuncs.h>
 
 #include <groonga.h>
@@ -21,35 +23,50 @@
 
 PG_MODULE_MAGIC;
 
-typedef struct GrnBuildStateData
+static bool PGrnIsLZ4Available;
+static relopt_kind PGrnReloptionKind;
+
+typedef struct PGrnOptions
+{
+	int32 vl_len_;
+	int tokenizerOffset;
+	int normalizerOffset;
+} PGrnOptions;
+
+typedef struct PGrnBuildStateData
 {
 	grn_obj	*idsTable;
 	double nIndexedTuples;
-} GrnBuildStateData;
+} PGrnBuildStateData;
 
-typedef GrnBuildStateData *GrnBuildState;
+typedef PGrnBuildStateData *PGrnBuildState;
 
-typedef struct GrnScanOpaqueData
+typedef struct PGrnScanOpaqueData
 {
 	grn_obj *idsTable;
+	grn_obj *lexicon;
+	grn_obj *indexColumn;
+	grn_obj minBorderValue;
+	grn_obj maxBorderValue;
 	grn_obj *searched;
 	grn_obj *sorted;
 	grn_obj *targetTable;
-	grn_table_cursor *cursor;
+	grn_obj *indexCursor;
+	grn_table_cursor *tableCursor;
 	grn_obj *keyAccessor;
 	grn_id currentID;
-} GrnScanOpaqueData;
+} PGrnScanOpaqueData;
 
-typedef GrnScanOpaqueData *GrnScanOpaque;
+typedef PGrnScanOpaqueData *PGrnScanOpaque;
 
-typedef struct GrnSearchData
+typedef struct PGrnSearchData
 {
 	grn_obj	*indexColumn;
 	grn_obj matchTargets;
 	grn_obj sectionID;
 	grn_obj *expression;
 	grn_obj *expressionVariable;
-} GrnSearchData;
+} PGrnSearchData;
 
 
 PG_FUNCTION_INFO_V1(pgroonga_contain_text);
@@ -74,36 +91,35 @@ static grn_ctx *ctx = &grnContext;
 static grn_obj buffer;
 static grn_obj inspectBuffer;
 
-#ifdef PGROONGA_DEBUG
 static const char *
-GrnInspect(grn_obj *object)
+PGrnInspect(grn_obj *object)
 {
 	GRN_BULK_REWIND(&inspectBuffer);
 	grn_inspect(ctx, &inspectBuffer, object);
 	GRN_TEXT_PUTC(ctx, &inspectBuffer, '\0');
 	return GRN_TEXT_VALUE(&inspectBuffer);
 }
-#endif
 
 static grn_encoding
-GrnGetEncoding(void)
+PGrnGetEncoding(void)
 {
 	int	enc = GetDatabaseEncoding();
 
-	if (pg_encoding_max_length(enc) > 1)
-		return GRN_ENC_NONE;
-
 	switch (enc)
 	{
+	case PG_SQL_ASCII:
+	case PG_UTF8:
+		return GRN_ENC_UTF8;
 	case PG_EUC_JP:
 	case PG_EUC_JIS_2004:
 		return GRN_ENC_EUC_JP;
-	case PG_UTF8:
-		return GRN_ENC_UTF8;
 	case PG_LATIN1:
 		return GRN_ENC_LATIN1;
 	case PG_KOI8R:
 		return GRN_ENC_KOI8R;
+	case PG_SJIS:
+	case PG_SHIFT_JIS_2004:
+		return GRN_ENC_SJIS;
 	default:
 		elog(WARNING,
 			 "pgroonga: use default encoding instead of '%s'",
@@ -113,15 +129,15 @@ GrnGetEncoding(void)
 }
 
 static void
-GrnEnsureDatabase(void)
+PGrnEnsureDatabase(void)
 {
 	char path[MAXPGPATH];
 	grn_obj	*db;
 
-	GRN_CTX_SET_ENCODING(ctx, GrnGetEncoding());
+	GRN_CTX_SET_ENCODING(ctx, PGrnGetEncoding());
 	join_path_components(path,
 						 GetDatabasePath(MyDatabaseId, DEFAULTTABLESPACE_OID),
-						 GrnDatabaseBasename);
+						 PGrnDatabaseBasename);
 
 	db = grn_db_open(ctx, path);
 	if (db)
@@ -136,7 +152,7 @@ GrnEnsureDatabase(void)
 }
 
 static void
-GrnOnProcExit(int code, Datum arg)
+PGrnOnProcExit(int code, Datum arg)
 {
 	grn_obj *db;
 
@@ -151,6 +167,114 @@ GrnOnProcExit(int code, Datum arg)
 	grn_fin();
 }
 
+static void
+PGrnInitializeGroongaInformation(void)
+{
+	grn_obj grnIsSupported;
+
+	GRN_BOOL_INIT(&grnIsSupported, 0);
+	grn_obj_get_info(ctx, NULL, GRN_INFO_SUPPORT_LZ4, &grnIsSupported);
+	PGrnIsLZ4Available = (GRN_BOOL_VALUE(&grnIsSupported));
+	GRN_OBJ_FIN(ctx, &grnIsSupported);
+}
+
+static bool
+PGrnIsTokenizer(grn_obj *object)
+{
+	if (object->header.type != GRN_PROC)
+		return false;
+
+  if (grn_proc_get_type(ctx, object) != GRN_PROC_TOKENIZER)
+	  return false;
+
+  return true;
+}
+
+static void
+PGrnOptionValidateTokenizer(char *name)
+{
+	grn_obj *tokenizer;
+	size_t name_length;
+
+	name_length = strlen(name);
+	if (name_length == 0)
+		return;
+
+	tokenizer = grn_ctx_get(ctx, name, name_length);
+	if (!tokenizer)
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("pgroonga: nonexistent tokenizer: <%s>",
+						name)));
+	}
+
+	if (!PGrnIsTokenizer(tokenizer))
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("pgroonga: not tokenizer: <%s>: %s",
+						name, PGrnInspect(tokenizer))));
+	}
+}
+
+static bool
+PGrnIsNormalizer(grn_obj *object)
+{
+	if (object->header.type != GRN_PROC)
+		return false;
+
+  if (grn_proc_get_type(ctx, object) != GRN_PROC_NORMALIZER)
+	  return false;
+
+  return true;
+}
+
+static void
+PGrnOptionValidateNormalizer(char *name)
+{
+	grn_obj *normalizer;
+	size_t name_length;
+
+	name_length = strlen(name);
+	if (name_length == 0)
+		return;
+
+	normalizer = grn_ctx_get(ctx, name, name_length);
+	if (!normalizer)
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("pgroonga: nonexistent normalizer: <%s>",
+						name)));
+	}
+
+	if (!PGrnIsNormalizer(normalizer))
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("pgroonga: not normalizer: <%s>: %s",
+						name, PGrnInspect(normalizer))));
+	}
+}
+
+static void
+PGrnInitializeOptions(void)
+{
+	PGrnReloptionKind = add_reloption_kind();
+
+	add_string_reloption(PGrnReloptionKind,
+						 "tokenizer",
+						 "Tokenizer name to be used for full-text search",
+						 PGRN_DEFAULT_TOKENIZER,
+						 PGrnOptionValidateTokenizer);
+	add_string_reloption(PGrnReloptionKind,
+						 "normalizer",
+						 "Normalizer name to be used for full-text search",
+						 PGRN_DEFAULT_NORMALIZER,
+						 PGrnOptionValidateNormalizer);
+}
+
 void
 _PG_init(void)
 {
@@ -163,16 +287,20 @@ _PG_init(void)
 				(errcode(ERRCODE_SYSTEM_ERROR),
 				 errmsg("pgroonga: failed to initialize Groonga context")));
 
-	on_proc_exit(GrnOnProcExit, 0);
+	on_proc_exit(PGrnOnProcExit, 0);
 
 	GRN_VOID_INIT(&buffer);
 	GRN_TEXT_INIT(&inspectBuffer, 0);
 
-	GrnEnsureDatabase();
+	PGrnEnsureDatabase();
+
+	PGrnInitializeGroongaInformation();
+
+	PGrnInitializeOptions();
 }
 
 static int
-GrnRCToPgErrorCode(grn_rc rc)
+PGrnRCToPgErrorCode(grn_rc rc)
 {
 	int errorCode = ERRCODE_SYSTEM_ERROR;
 
@@ -196,13 +324,13 @@ GrnRCToPgErrorCode(grn_rc rc)
 }
 
 static grn_bool
-GrnCheck(const char *message)
+PGrnCheck(const char *message)
 {
 	if (ctx->rc == GRN_SUCCESS)
 		return GRN_TRUE;
 
 	ereport(ERROR,
-			(errcode(GrnRCToPgErrorCode(ctx->rc)),
+			(errcode(PGrnRCToPgErrorCode(ctx->rc)),
 			 errmsg("pgroonga: %s: %s", message, ctx->errbuf)));
 	return GRN_FALSE;
 }
@@ -211,33 +339,33 @@ GrnCheck(const char *message)
  * Support functions and type-specific routines
  */
 
-static grn_builtin_type
-GrnGetType(Relation index, AttrNumber n)
+static grn_id
+PGrnGetType(Relation index, AttrNumber n)
 {
 	FmgrInfo *function;
 	TupleDesc desc = RelationGetDescr(index);
 	Datum type;
 
-	function = index_getprocinfo(index, n + 1, GrnTypeOfProc);
+	function = index_getprocinfo(index, n + 1, PGrnTypeOfProc);
 	type = FunctionCall2(function,
 						 ObjectIdGetDatum(desc->attrs[n]->atttypid),
 						 Int32GetDatum(desc->attrs[n]->atttypmod));
-	return (grn_builtin_type) DatumGetInt32(type);
+	return DatumGetInt32(type);
 }
 
 static void
-GrnGetValue(Relation index, AttrNumber n, grn_obj *buffer, Datum value)
+PGrnGetValue(Relation index, AttrNumber n, grn_obj *buffer, Datum value)
 {
 	FmgrInfo *function;
 
-	function = index_getprocinfo(index, n + 1, GrnGetValueProc);
+	function = index_getprocinfo(index, n + 1, PGrnGetValueProc);
 	FunctionCall3(function,
 				  PointerGetDatum(ctx), PointerGetDatum(buffer),
 				  value);
 }
 
 static grn_obj *
-GrnLookup(const char *name, int errorLevel)
+PGrnLookup(const char *name, int errorLevel)
 {
 	grn_obj *object = grn_ctx_get(ctx, name, strlen(name));
 	if (!object)
@@ -248,27 +376,27 @@ GrnLookup(const char *name, int errorLevel)
 }
 
 static grn_obj *
-GrnLookupIDsTable(Relation index, int errorLevel)
+PGrnLookupIDsTable(Relation index, int errorLevel)
 {
 	char name[GRN_TABLE_MAX_KEY_SIZE];
 
-	snprintf(name, sizeof(name), GrnIDsTableNameFormat, index->rd_node.relNode);
-	return GrnLookup(name, errorLevel);
+	snprintf(name, sizeof(name), PGrnIDsTableNameFormat, index->rd_node.relNode);
+	return PGrnLookup(name, errorLevel);
 }
 
 static grn_obj *
-GrnLookupIndexColumn(Relation index, int errorLevel)
+PGrnLookupIndexColumn(Relation index, int errorLevel)
 {
 	char name[GRN_TABLE_MAX_KEY_SIZE];
 
 	snprintf(name, sizeof(name),
-			 GrnLexiconNameFormat ".%s",
-			 index->rd_node.relNode, GrnIndexColumnName);
-	return GrnLookup(name, errorLevel);
+			 PGrnLexiconNameFormat ".%s",
+			 index->rd_node.relNode, PGrnIndexColumnName);
+	return PGrnLookup(name, errorLevel);
 }
 
 static grn_obj *
-GrnCreateTable(const char *name,
+PGrnCreateTable(const char *name,
 			   grn_obj_flags flags,
 			   grn_obj *type)
 {
@@ -279,13 +407,13 @@ GrnCreateTable(const char *name,
 							 GRN_OBJ_PERSISTENT | flags,
 							 type,
 							 NULL);
-	GrnCheck("pgroonga: failed to create table");
+	PGrnCheck("pgroonga: failed to create table");
 
 	return table;
 }
 
 static grn_obj *
-GrnCreateColumn(grn_obj	*table,
+PGrnCreateColumn(grn_obj	*table,
 				const char *name,
 				grn_obj_flags flags,
 				grn_obj	*type)
@@ -296,20 +424,31 @@ GrnCreateColumn(grn_obj	*table,
 							   name, strlen(name), NULL,
 							   GRN_OBJ_PERSISTENT | flags,
 							   type);
-	GrnCheck("pgroonga: failed to create column");
+	PGrnCheck("pgroonga: failed to create column");
 
 	return column;
 }
 
+static bool
+PGrnIsForFullTextSearchIndex(Relation index)
+{
+	Oid containStrategyOID;
+	containStrategyOID = get_opfamily_member(index->rd_opfamily[0],
+											 index->rd_opcintype[0],
+											 index->rd_opcintype[0],
+											 PGrnContainStrategyNumber);
+	return (containStrategyOID != InvalidOid);
+}
+
 /**
- * GrnCreate
+ * PGrnCreate
  *
  * @param	ctx
  * @param	index
  */
 static void
-GrnCreate(Relation index, grn_obj **idsTable,
-		  grn_obj **lexicon, grn_obj **indexColumn)
+PGrnCreate(Relation index, grn_obj **idsTable,
+		   grn_obj **lexicon, grn_obj **indexColumn)
 {
 	char idsTableName[GRN_TABLE_MAX_KEY_SIZE];
 	char lexiconName[GRN_TABLE_MAX_KEY_SIZE];
@@ -317,20 +456,22 @@ GrnCreate(Relation index, grn_obj **idsTable,
 	int i;
 	TupleDesc desc;
 	Oid relNode = index->rd_node.relNode;
+	bool forFullTextSearch = false;
 
 	desc = RelationGetDescr(index);
 
 	snprintf(idsTableName, sizeof(idsTableName),
-			 GrnIDsTableNameFormat, relNode);
-	*idsTable = GrnCreateTable(idsTableName,
-							   GRN_OBJ_TABLE_PAT_KEY,
-							   grn_ctx_at(ctx, GRN_DB_UINT64));
+			 PGrnIDsTableNameFormat, relNode);
+	*idsTable = PGrnCreateTable(idsTableName,
+								GRN_OBJ_TABLE_PAT_KEY,
+								grn_ctx_at(ctx, GRN_DB_UINT64));
 
 	for (i = 0; i < desc->natts; i++)
 	{
 		grn_id attributeTypeID;
+		grn_obj_flags flags = GRN_OBJ_COLUMN_SCALAR;
 
-		attributeTypeID = GrnGetType(index, i);
+		attributeTypeID = PGrnGetType(index, i);
 		if (typeID == GRN_ID_NIL)
 			typeID = attributeTypeID;
 
@@ -343,12 +484,26 @@ GrnCreate(Relation index, grn_obj **idsTable,
 							"for multiple column index")));
 		}
 
-		GrnCreateColumn(*idsTable,
-						desc->attrs[i]->attname.data,
-						GRN_OBJ_COLUMN_SCALAR,
-						grn_ctx_at(ctx, attributeTypeID));
+		if (PGrnIsLZ4Available)
+		{
+			switch (typeID)
+			{
+			case GRN_DB_SHORT_TEXT:
+			case GRN_DB_TEXT:
+			case GRN_DB_LONG_TEXT:
+				flags |= GRN_OBJ_COMPRESS_LZ4;
+				break;
+			}
+		}
+
+		PGrnCreateColumn(*idsTable,
+						 desc->attrs[i]->attname.data,
+						 flags,
+						 grn_ctx_at(ctx, attributeTypeID));
 	}
 
+	forFullTextSearch = PGrnIsForFullTextSearchIndex(index);
+
 	switch (typeID)
 	{
 	case GRN_DB_TEXT:
@@ -357,33 +512,49 @@ GrnCreate(Relation index, grn_obj **idsTable,
 		break;
 	}
 
-	snprintf(lexiconName, sizeof(lexiconName), GrnLexiconNameFormat, relNode);
-	*lexicon = GrnCreateTable(lexiconName,
-							  GRN_OBJ_TABLE_PAT_KEY,
-							  grn_ctx_at(ctx, typeID));
-	if (typeID == GRN_DB_SHORT_TEXT)
+	snprintf(lexiconName, sizeof(lexiconName), PGrnLexiconNameFormat, relNode);
+	*lexicon = PGrnCreateTable(lexiconName,
+							   GRN_OBJ_TABLE_PAT_KEY,
+							   grn_ctx_at(ctx, typeID));
+	if (forFullTextSearch)
 	{
-		grn_obj_set_info(ctx, *lexicon, GRN_INFO_NORMALIZER,
-						 GrnLookup("NormalizerAuto", WARNING));
-		grn_obj_set_info(ctx, *lexicon, GRN_INFO_DEFAULT_TOKENIZER,
-						 grn_ctx_at(ctx, GRN_DB_BIGRAM));
+		PGrnOptions *options;
+		const char *tokenizerName = PGRN_DEFAULT_TOKENIZER;
+		const char *normalizerName = PGRN_DEFAULT_NORMALIZER;
+
+		options = (PGrnOptions *) (index->rd_options);
+		if (options)
+		{
+			tokenizerName = (const char *) (options) + options->tokenizerOffset;
+			normalizerName = (const char *) (options) + options->normalizerOffset;
+		}
+		if (tokenizerName && tokenizerName[0])
+		{
+			grn_obj_set_info(ctx, *lexicon, GRN_INFO_DEFAULT_TOKENIZER,
+							 PGrnLookup(tokenizerName, ERROR));
+		}
+		if (normalizerName && normalizerName[0])
+		{
+			grn_obj_set_info(ctx, *lexicon, GRN_INFO_NORMALIZER,
+							 PGrnLookup(normalizerName, ERROR));
+		}
 	}
 
 	{
 		grn_obj_flags flags = GRN_OBJ_COLUMN_INDEX;
-		if (typeID == GRN_DB_SHORT_TEXT)
+		if (forFullTextSearch)
 			flags |= GRN_OBJ_WITH_POSITION;
 		if (desc->natts > 1)
 			flags |= GRN_OBJ_WITH_SECTION;
-		*indexColumn = GrnCreateColumn(*lexicon,
-									   GrnIndexColumnName,
-									   flags,
-									   *idsTable);
+		*indexColumn = PGrnCreateColumn(*lexicon,
+										PGrnIndexColumnName,
+										flags,
+										*idsTable);
 	}
 }
 
 static void
-GrnSetSources(Relation index, grn_obj *idsTable, grn_obj *indexColumn)
+PGrnSetSources(Relation index, grn_obj *idsTable, grn_obj *indexColumn)
 {
 	TupleDesc desc;
 	grn_obj source_ids;
@@ -426,26 +597,6 @@ UInt64ToCtid(uint64 key)
 	return ctid;
 }
 
-static void
-GrnLock(Relation index, LOCKMODE mode)
-{
-	const RelFileNode *rnode = &index->rd_node;
-	LockDatabaseObject(rnode->spcNode,
-					   rnode->dbNode,
-					   rnode->relNode,
-					   mode);
-}
-
-static void
-GrnUnlock(Relation index, LOCKMODE mode)
-{
-	const RelFileNode *rnode = &index->rd_node;
-	UnlockDatabaseObject(rnode->spcNode,
-						 rnode->dbNode,
-						 rnode->relNode,
-						 mode);
-}
-
 static grn_bool
 pgroonga_contain_raw(const char *text, unsigned int text_size,
 					 const char *key, unsigned int key_size)
@@ -533,7 +684,7 @@ pgroonga_match(PG_FUNCTION_ARGS)
 }
 
 static void
-GrnInsert(grn_ctx *ctx,
+PGrnInsert(grn_ctx *ctx,
 		  Relation index,
 		  grn_obj *idsTable,
 		  Datum *values,
@@ -557,11 +708,11 @@ GrnInsert(grn_ctx *ctx,
 
 		dataColumn = grn_obj_column(ctx, idsTable,
 									name->data, strlen(name->data));
-		grn_obj_reinit(ctx, &buffer, GrnGetType(index, i), 0);
-		GrnGetValue(index, i, &buffer, values[i]);
+		grn_obj_reinit(ctx, &buffer, PGrnGetType(index, i), 0);
+		PGrnGetValue(index, i, &buffer, values[i]);
 		grn_obj_set_value(ctx, dataColumn, id, &buffer, GRN_OBJ_SET);
 		grn_obj_unlink(ctx, dataColumn);
-		if (!GrnCheck("pgroonga: failed to set column value")) {
+		if (!PGrnCheck("pgroonga: failed to set column value")) {
 			continue;
 		}
 	}
@@ -581,29 +732,32 @@ pgroonga_insert(PG_FUNCTION_ARGS)
 	Relation heap = (Relation) PG_GETARG_POINTER(4);
 	IndexUniqueCheck checkUnique = PG_GETARG_INT32(5);
 #endif
-	grn_obj *idsTable = GrnLookupIDsTable(index, ERROR);
+	grn_obj *idsTable = PGrnLookupIDsTable(index, ERROR);
 
-	GrnLock(index, ExclusiveLock);
-	GrnInsert(ctx, index, idsTable, values, isnull, ht_ctid);
-	GrnUnlock(index, ExclusiveLock);
+	PGrnInsert(ctx, index, idsTable, values, isnull, ht_ctid);
 
 	PG_RETURN_BOOL(true);
 }
 
 static void
-GrnScanOpaqueInit(GrnScanOpaque so, Relation index)
+PGrnScanOpaqueInit(PGrnScanOpaque so, Relation index)
 {
-	so->idsTable = GrnLookupIDsTable(index, ERROR);
+	so->idsTable = PGrnLookupIDsTable(index, ERROR);
+	so->indexColumn = PGrnLookupIndexColumn(index, ERROR);
+	so->lexicon = grn_column_table(ctx, so->indexColumn);
+	GRN_VOID_INIT(&(so->minBorderValue));
+	GRN_VOID_INIT(&(so->maxBorderValue));
 	so->searched = NULL;
 	so->sorted = NULL;
 	so->targetTable = NULL;
-	so->cursor = NULL;
+	so->indexCursor = NULL;
+	so->tableCursor = NULL;
 	so->keyAccessor = NULL;
 	so->currentID = GRN_ID_NIL;
 }
 
 static void
-GrnScanOpaqueReinit(GrnScanOpaque so)
+PGrnScanOpaqueReinit(PGrnScanOpaque so)
 {
 	so->currentID = GRN_ID_NIL;
 	if (so->keyAccessor)
@@ -611,11 +765,18 @@ GrnScanOpaqueReinit(GrnScanOpaque so)
 		grn_obj_unlink(ctx, so->keyAccessor);
 		so->keyAccessor = NULL;
 	}
-	if (so->cursor)
+	if (so->indexCursor)
 	{
-		grn_table_cursor_close(ctx, so->cursor);
-		so->cursor = NULL;
+		grn_obj_close(ctx, so->indexCursor);
+		so->indexCursor = NULL;
 	}
+	if (so->tableCursor)
+	{
+		grn_table_cursor_close(ctx, so->tableCursor);
+		so->tableCursor = NULL;
+	}
+	GRN_OBJ_FIN(ctx, &(so->minBorderValue));
+	GRN_OBJ_FIN(ctx, &(so->maxBorderValue));
 	if (so->sorted)
 	{
 		grn_obj_unlink(ctx, so->sorted);
@@ -638,12 +799,12 @@ pgroonga_beginscan(PG_FUNCTION_ARGS)
 	int nkeys = PG_GETARG_INT32(1);
 	int norderbys = PG_GETARG_INT32(2);
 	IndexScanDesc scan;
-	GrnScanOpaque so;
+	PGrnScanOpaque so;
 
 	scan = RelationGetIndexScan(index, nkeys, norderbys);
 
-	so = (GrnScanOpaque) palloc(sizeof(GrnScanOpaqueData));
-	GrnScanOpaqueInit(so, index);
+	so = (PGrnScanOpaque) palloc(sizeof(PGrnScanOpaqueData));
+	PGrnScanOpaqueInit(so, index);
 
 	scan->opaque = so;
 
@@ -651,9 +812,9 @@ pgroonga_beginscan(PG_FUNCTION_ARGS)
 }
 
 static void
-GrnSearchBuildConditions(IndexScanDesc scan,
-						 GrnScanOpaque so,
-						 GrnSearchData *data)
+PGrnSearchBuildConditions(IndexScanDesc scan,
+						  PGrnScanOpaque so,
+						  PGrnSearchData *data)
 {
 	Relation index = scan->indexRelation;
 	int i, nExpressions = 0;
@@ -682,33 +843,30 @@ GrnSearchBuildConditions(IndexScanDesc scan,
 
 		grn_expr_append_op(ctx, matchTarget, GRN_OP_GET_MEMBER, 2);
 
-		grn_obj_reinit(ctx, &buffer, GrnGetType(index, key->sk_attno - 1), 0);
-		GrnGetValue(index, key->sk_attno - 1, &buffer, key->sk_argument);
+		grn_obj_reinit(ctx, &buffer, PGrnGetType(index, key->sk_attno - 1), 0);
+		PGrnGetValue(index, key->sk_attno - 1, &buffer, key->sk_argument);
 
 		switch (key->sk_strategy)
 		{
-		case GrnLessStrategyNumber:
+		case PGrnLessStrategyNumber:
 			operator = GRN_OP_LESS;
 			break;
-		case GrnLessEqualStrategyNumber:
+		case PGrnLessEqualStrategyNumber:
 			operator = GRN_OP_LESS_EQUAL;
 			break;
-		case GrnEqualStrategyNumber:
+		case PGrnEqualStrategyNumber:
 			operator = GRN_OP_EQUAL;
 			break;
-		case GrnGreaterEqualStrategyNumber:
+		case PGrnGreaterEqualStrategyNumber:
 			operator = GRN_OP_GREATER_EQUAL;
 			break;
-		case GrnGreaterStrategyNumber:
+		case PGrnGreaterStrategyNumber:
 			operator = GRN_OP_GREATER;
 			break;
-		case GrnNotEqualStrategyNumber:
-			operator = GRN_OP_NOT_EQUAL;
-			break;
-		case GrnContainStrategyNumber:
+		case PGrnContainStrategyNumber:
 			operator = GRN_OP_MATCH;
 			break;
-		case GrnQueryStrategyNumber:
+		case PGrnQueryStrategyNumber:
 			break;
 		default:
 			ereport(ERROR,
@@ -722,7 +880,7 @@ GrnSearchBuildConditions(IndexScanDesc scan,
 		if (!isValidStrategy)
 			continue;
 
-		if (key->sk_strategy == GrnQueryStrategyNumber)
+		if (key->sk_strategy == PGrnQueryStrategyNumber)
 		{
 			grn_rc rc;
 			grn_expr_flags flags =
@@ -734,7 +892,7 @@ GrnSearchBuildConditions(IndexScanDesc scan,
 			if (rc != GRN_SUCCESS)
 			{
 				ereport(ERROR,
-						(errcode(GrnRCToPgErrorCode(rc)),
+						(errcode(PGrnRCToPgErrorCode(rc)),
 						 errmsg("pgroonga: failed to parse expression: %s",
 								ctx->errbuf)));
 			}
@@ -755,7 +913,7 @@ GrnSearchBuildConditions(IndexScanDesc scan,
 }
 
 static void
-GrnSearchDataFree(GrnSearchData *data)
+PGrnSearchDataFree(PGrnSearchData *data)
 {
 	unsigned int i, nMatchTargets;
 
@@ -771,11 +929,11 @@ GrnSearchDataFree(GrnSearchData *data)
 }
 
 static void
-GrnSearch(IndexScanDesc scan)
+PGrnSearch(IndexScanDesc scan)
 {
 	Relation index = scan->indexRelation;
-	GrnScanOpaque so = (GrnScanOpaque) scan->opaque;
-	GrnSearchData data;
+	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
+	PGrnSearchData data;
 
 	if (scan->numberOfKeys == 0)
 		return;
@@ -783,18 +941,18 @@ GrnSearch(IndexScanDesc scan)
 	GRN_PTR_INIT(&(data.matchTargets), GRN_OBJ_VECTOR, GRN_ID_NIL);
 	GRN_UINT32_INIT(&(data.sectionID), 0);
 
-	data.indexColumn = GrnLookupIndexColumn(index, ERROR);
+	data.indexColumn = PGrnLookupIndexColumn(index, ERROR);
 
 	GRN_EXPR_CREATE_FOR_QUERY(ctx, so->idsTable,
 							  data.expression, data.expressionVariable);
 
 	PG_TRY();
 	{
-		GrnSearchBuildConditions(scan, so, &data);
+		PGrnSearchBuildConditions(scan, so, &data);
 	}
 	PG_CATCH();
 	{
-		GrnSearchDataFree(&data);
+		PGrnSearchDataFree(&data);
 		PG_RE_THROW();
 	}
 	PG_END_TRY();
@@ -808,19 +966,19 @@ GrnSearch(IndexScanDesc scan)
 					 data.expression,
 					 so->searched,
 					 GRN_OP_OR);
-	GrnSearchDataFree(&data);
+	PGrnSearchDataFree(&data);
 }
 
 static void
-GrnSort(IndexScanDesc scan)
+PGrnSort(IndexScanDesc scan)
 {
 	/* TODO */
 }
 
 static void
-GrnOpenCursor(IndexScanDesc scan, ScanDirection dir)
+PGrnOpenTableCursor(IndexScanDesc scan, ScanDirection dir)
 {
-	GrnScanOpaque so = (GrnScanOpaque) scan->opaque;
+	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
 	grn_obj *table;
 	int offset = 0;
 	int limit = -1;
@@ -837,25 +995,206 @@ GrnOpenCursor(IndexScanDesc scan, ScanDirection dir)
 	else
 		flags |= GRN_CURSOR_ASCENDING;
 
-	so->cursor = grn_table_cursor_open(ctx, table,
-									   NULL, 0, NULL, 0,
-									   offset, limit, flags);
+	so->tableCursor = grn_table_cursor_open(ctx, table,
+											NULL, 0, NULL, 0,
+											offset, limit, flags);
 	so->keyAccessor = grn_obj_column(ctx, table,
 									 GRN_COLUMN_NAME_KEY,
 									 GRN_COLUMN_NAME_KEY_LEN);
 }
 
-static void
-GrnEnsureCursorOpened(IndexScanDesc scan, ScanDirection dir)
+static bool
+PGrnIsMeaningfullMaxBorderValue(grn_obj *currentValue,
+								grn_obj *newValue,
+								int flags,
+								StrategyNumber strategy)
 {
-	GrnScanOpaque so = (GrnScanOpaque) scan->opaque;
+	if (((flags & GRN_CURSOR_LT) == GRN_CURSOR_LT) &&
+		strategy == PGrnLessEqualStrategyNumber)
+	{
+		return grn_operator_exec_greater_equal(ctx, currentValue, newValue);
+	}
+	else
+	{
+		return grn_operator_exec_greater(ctx, currentValue, newValue);
+	}
+}
 
-	if (so->cursor)
+static bool
+PGrnIsMeaningfullMinBorderValue(grn_obj *currentValue,
+								grn_obj *newValue,
+								int flags,
+								StrategyNumber strategy)
+{
+	if (((flags & GRN_CURSOR_GT) == GRN_CURSOR_GT) &&
+		strategy == PGrnGreaterEqualStrategyNumber)
+	{
+		return grn_operator_exec_less_equal(ctx, currentValue, newValue);
+	}
+	else
+	{
+		return grn_operator_exec_less(ctx, currentValue, newValue);
+	}
+}
+
+static void
+PGrnFillBorder(IndexScanDesc scan,
+			   void **min, unsigned int *minSize,
+			   void **max, unsigned int *maxSize,
+			   int *flags)
+{
+	Relation index = scan->indexRelation;
+	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
+	grn_obj *minBorderValue;
+	grn_obj *maxBorderValue;
+	int i;
+
+	minBorderValue = &(so->minBorderValue);
+	maxBorderValue = &(so->maxBorderValue);
+	for (i = 0; i < scan->numberOfKeys; i++)
+	{
+		ScanKey key = &(scan->keyData[i]);
+		AttrNumber attrNumber;
+		grn_id domain;
+
+		attrNumber = key->sk_attno - 1;
+		domain = PGrnGetType(index, attrNumber);
+		switch (key->sk_strategy)
+		{
+		case PGrnLessStrategyNumber:
+		case PGrnLessEqualStrategyNumber:
+			if (maxBorderValue->header.type != GRN_DB_VOID)
+			{
+				grn_obj_reinit(ctx, &buffer, domain, 0);
+				PGrnGetValue(index, attrNumber, &buffer, key->sk_argument);
+				if (!PGrnIsMeaningfullMaxBorderValue(maxBorderValue,
+													 &buffer,
+													 *flags,
+													 key->sk_strategy))
+				{
+					continue;
+				}
+			}
+			grn_obj_reinit(ctx, maxBorderValue, domain, 0);
+			PGrnGetValue(index, attrNumber, maxBorderValue, key->sk_argument);
+			*max = GRN_BULK_HEAD(maxBorderValue);
+			*maxSize = GRN_BULK_VSIZE(maxBorderValue);
+			*flags &= ~(GRN_CURSOR_LT | GRN_CURSOR_LE);
+			if (key->sk_strategy == PGrnLessStrategyNumber)
+			{
+				*flags |= GRN_CURSOR_LT;
+			}
+			else
+			{
+				*flags |= GRN_CURSOR_LE;
+			}
+			break;
+		case PGrnEqualStrategyNumber:
+			/* TODO: Support the following:
+			 *   WHERE id  = 5 AND id <= 4 (Empty result)
+			 *   WHERE id  = 2 AND id >= 4 (Empty result)
+			 *   WHERE id <= 4 AND id  = 5 (Empty result)
+			 *   WHERE id >= 4 AND id  = 2 (Empty result)
+			 *   WHERE id  = 2 AND id <= 4 (Ignore id = 2)
+			 *   WHERE id  = 4 AND id >= 2 (Ignore id = 4)
+			 */
+			grn_obj_reinit(ctx, minBorderValue, domain, 0);
+			PGrnGetValue(index, attrNumber, minBorderValue, key->sk_argument);
+			grn_obj_reinit(ctx, maxBorderValue, domain, 0);
+			PGrnGetValue(index, attrNumber, maxBorderValue, key->sk_argument);
+			*flags |= GRN_CURSOR_LE | GRN_CURSOR_GE;
+			break;
+		case PGrnGreaterEqualStrategyNumber:
+		case PGrnGreaterStrategyNumber:
+			if (minBorderValue->header.type != GRN_DB_VOID)
+			{
+				grn_obj_reinit(ctx, &buffer, domain, 0);
+				PGrnGetValue(index, attrNumber, &buffer,
+							 key->sk_argument);
+				if (!PGrnIsMeaningfullMinBorderValue(minBorderValue,
+													 &buffer,
+													 *flags,
+													 key->sk_strategy))
+				{
+					continue;
+				}
+			}
+			grn_obj_reinit(ctx, minBorderValue, domain, 0);
+			PGrnGetValue(index, attrNumber, minBorderValue, key->sk_argument);
+			*min = GRN_BULK_HEAD(minBorderValue);
+			*minSize = GRN_BULK_VSIZE(minBorderValue);
+			*flags &= ~(GRN_CURSOR_GT | GRN_CURSOR_GE);
+			if (key->sk_strategy == PGrnGreaterEqualStrategyNumber)
+			{
+				*flags |= GRN_CURSOR_GE;
+			}
+			else
+			{
+				*flags |= GRN_CURSOR_GT;
+			}
+			break;
+		}
+	}
+}
+
+static void
+PGrnOpenIndexCursor(IndexScanDesc scan, ScanDirection dir)
+{
+	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
+	void *min = NULL;
+	unsigned int minSize = 0;
+	void *max = NULL;
+	unsigned int maxSize = 0;
+	int offset = 0;
+	int limit = -1;
+	int flags = 0;
+	grn_id indexCursorMin = GRN_ID_NIL;
+	grn_id indexCursorMax = GRN_ID_MAX;
+	int indexCursorFlags = 0;
+
+	PGrnFillBorder(scan, &min, &minSize, &max, &maxSize, &flags);
+
+	if (dir == BackwardScanDirection)
+		flags |= GRN_CURSOR_DESCENDING;
+	else
+		flags |= GRN_CURSOR_ASCENDING;
+
+	so->tableCursor = grn_table_cursor_open(ctx, so->lexicon,
+											min, minSize,
+											max, maxSize,
+											offset, limit, flags);
+	so->indexCursor = grn_index_cursor_open(ctx,
+											so->tableCursor, so->indexColumn,
+											indexCursorMin,
+											indexCursorMax,
+											indexCursorFlags);
+	so->keyAccessor = grn_obj_column(ctx, so->idsTable,
+									 GRN_COLUMN_NAME_KEY,
+									 GRN_COLUMN_NAME_KEY_LEN);
+}
+
+static void
+PGrnEnsureCursorOpened(IndexScanDesc scan, ScanDirection dir)
+{
+	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
+	bool forFullTextSearch = false;
+
+	if (so->indexCursor)
+		return;
+	if (so->tableCursor)
 		return;
 
-	GrnSearch(scan);
-	GrnSort(scan);
-	GrnOpenCursor(scan, dir);
+	forFullTextSearch = PGrnIsForFullTextSearchIndex(scan->indexRelation);
+	if (forFullTextSearch)
+	{
+		PGrnSearch(scan);
+		PGrnSort(scan);
+		PGrnOpenTableCursor(scan, dir);
+	}
+	else
+	{
+		PGrnOpenIndexCursor(scan, dir);
+	}
 }
 
 
@@ -867,25 +1206,37 @@ pgroonga_gettuple(PG_FUNCTION_ARGS)
 {
 	IndexScanDesc scan = (IndexScanDesc) PG_GETARG_POINTER(0);
 	ScanDirection dir = (ScanDirection) PG_GETARG_INT32(1);
-	GrnScanOpaque so = (GrnScanOpaque) scan->opaque;
+	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
 
 	scan->xs_recheck = false;
 
-	GrnEnsureCursorOpened(scan, dir);
+	PGrnEnsureCursorOpened(scan, dir);
 
 	if (scan->kill_prior_tuple && so->currentID != GRN_ID_NIL)
 	{
 		grn_obj key;
 		GRN_UINT64_INIT(&key, 0);
-		GrnLock(scan->indexRelation, ExclusiveLock);
 		grn_obj_get_value(ctx, so->keyAccessor, so->currentID, &key);
 		grn_table_delete(ctx, so->idsTable,
 						 GRN_BULK_HEAD(&key), GRN_BULK_VSIZE(&key));
-		GrnUnlock(scan->indexRelation, ExclusiveLock);
 		GRN_OBJ_FIN(ctx, &key);
 	}
 
-	so->currentID = grn_table_cursor_next(ctx, so->cursor);
+	if (so->indexCursor)
+	{
+		grn_posting *posting;
+		grn_id termID;
+		grn_id id = GRN_ID_NIL;
+		posting = grn_index_cursor_next(ctx, so->indexCursor, &termID);
+		if (posting)
+			id = posting->rid;
+		so->currentID = id;
+	}
+	else
+	{
+		so->currentID = grn_table_cursor_next(ctx, so->tableCursor);
+	}
+
 	if (so->currentID == GRN_ID_NIL)
 	{
 		PG_RETURN_BOOL(false);
@@ -912,21 +1263,39 @@ pgroonga_getbitmap(PG_FUNCTION_ARGS)
 {
 	IndexScanDesc scan = (IndexScanDesc) PG_GETARG_POINTER(0);
 	TIDBitmap *tbm = (TIDBitmap *) PG_GETARG_POINTER(1);
-	GrnScanOpaque so = (GrnScanOpaque) scan->opaque;
+	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
 	int64 nRecords = 0;
-	grn_id id;
 	grn_obj key;
 
-	GrnEnsureCursorOpened(scan, ForwardScanDirection);
+	PGrnEnsureCursorOpened(scan, ForwardScanDirection);
 
 	GRN_UINT64_INIT(&key, 0);
-	while ((id = grn_table_cursor_next(ctx, so->cursor)) != GRN_ID_NIL) {
-		ItemPointerData ctid;
-		GRN_BULK_REWIND(&key);
-		grn_obj_get_value(ctx, so->keyAccessor, id, &key);
-		ctid = UInt64ToCtid(GRN_UINT64_VALUE(&key));
-		tbm_add_tuples(tbm, &ctid, 1, false);
-		nRecords++;
+	if (so->indexCursor)
+	{
+		grn_posting *posting;
+		grn_id termID;
+		while ((posting = grn_index_cursor_next(ctx, so->indexCursor, &termID)))
+		{
+			ItemPointerData ctid;
+			GRN_BULK_REWIND(&key);
+			grn_obj_get_value(ctx, so->keyAccessor, posting->rid, &key);
+			ctid = UInt64ToCtid(GRN_UINT64_VALUE(&key));
+			tbm_add_tuples(tbm, &ctid, 1, false);
+			nRecords++;
+		}
+	}
+	else
+	{
+		grn_id id;
+		while ((id = grn_table_cursor_next(ctx, so->tableCursor)) != GRN_ID_NIL)
+		{
+			ItemPointerData ctid;
+			GRN_BULK_REWIND(&key);
+			grn_obj_get_value(ctx, so->keyAccessor, id, &key);
+			ctid = UInt64ToCtid(GRN_UINT64_VALUE(&key));
+			tbm_add_tuples(tbm, &ctid, 1, false);
+			nRecords++;
+		}
 	}
 	GRN_OBJ_FIN(ctx, &key);
 
@@ -949,9 +1318,9 @@ pgroonga_rescan(PG_FUNCTION_ARGS)
 	ScanKey	orderbys = (ScanKey) PG_GETARG_POINTER(3);
 	int norderbys = PG_GETARG_INT32(4);
 #endif
-	GrnScanOpaque so = (GrnScanOpaque) scan->opaque;
+	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
 
-	GrnScanOpaqueReinit(so);
+	PGrnScanOpaqueReinit(so);
 
 	if (keys && scan->numberOfKeys > 0)
 		memmove(scan->keyData, keys, scan->numberOfKeys * sizeof(ScanKeyData));
@@ -966,26 +1335,26 @@ Datum
 pgroonga_endscan(PG_FUNCTION_ARGS)
 {
 	IndexScanDesc scan = (IndexScanDesc) PG_GETARG_POINTER(0);
-	GrnScanOpaque so = (GrnScanOpaque) scan->opaque;
+	PGrnScanOpaque so = (PGrnScanOpaque) scan->opaque;
 
-	GrnScanOpaqueReinit(so);
+	PGrnScanOpaqueReinit(so);
 	pfree(so);
 
 	PG_RETURN_VOID();
 }
 
 static void
-GrnBuildCallback(Relation index,
-				 HeapTuple htup,
-				 Datum *values,
-				 bool *isnull,
-				 bool tupleIsAlive,
-				 void *state)
+PGrnBuildCallback(Relation index,
+				  HeapTuple htup,
+				  Datum *values,
+				  bool *isnull,
+				  bool tupleIsAlive,
+				  void *state)
 {
-	GrnBuildState bs = (GrnBuildState) state;
+	PGrnBuildState bs = (PGrnBuildState) state;
 
 	if (tupleIsAlive) {
-		GrnInsert(ctx, index, bs->idsTable, values, isnull, &(htup->t_self));
+		PGrnInsert(ctx, index, bs->idsTable, values, isnull, &(htup->t_self));
 		bs->nIndexedTuples++;
 	}
 }
@@ -1001,7 +1370,7 @@ pgroonga_build(PG_FUNCTION_ARGS)
 	IndexInfo *indexInfo = (IndexInfo *) PG_GETARG_POINTER(2);
 	IndexBuildResult *result;
 	double nHeapTuples = 0.0;
-	GrnBuildStateData bs;
+	PGrnBuildStateData bs;
 	grn_obj *lexicon = NULL;
 	grn_obj *indexColumn = NULL;
 
@@ -1015,10 +1384,10 @@ pgroonga_build(PG_FUNCTION_ARGS)
 
 	PG_TRY();
 	{
-		GrnCreate(index, &(bs.idsTable), &lexicon, &indexColumn);
+		PGrnCreate(index, &(bs.idsTable), &lexicon, &indexColumn);
 		nHeapTuples = IndexBuildHeapScan(heap, index, indexInfo, true,
-										 GrnBuildCallback, &bs);
-		GrnSetSources(index, bs.idsTable, indexColumn);
+										 PGrnBuildCallback, &bs);
+		PGrnSetSources(index, bs.idsTable, indexColumn);
 	}
 	PG_CATCH();
 	{
@@ -1050,8 +1419,8 @@ pgroonga_buildempty(PG_FUNCTION_ARGS)
 
 	PG_TRY();
 	{
-		GrnCreate(index, &idsTable, &lexicon, &indexColumn);
-		GrnSetSources(index, idsTable, indexColumn);
+		PGrnCreate(index, &idsTable, &lexicon, &indexColumn);
+		PGrnSetSources(index, idsTable, indexColumn);
 	}
 	PG_CATCH();
 	{
@@ -1067,7 +1436,7 @@ pgroonga_buildempty(PG_FUNCTION_ARGS)
 }
 
 static IndexBulkDeleteResult *
-GrnBulkDeleteResult(IndexVacuumInfo *info, grn_obj *idsTable)
+PGrnBulkDeleteResult(IndexVacuumInfo *info, grn_obj *idsTable)
 {
 	IndexBulkDeleteResult *stats;
 
@@ -1098,10 +1467,10 @@ pgroonga_bulkdelete(PG_FUNCTION_ARGS)
 	grn_table_cursor *cursor;
 	double nRemovedTuples;
 
-	idsTable = GrnLookupIDsTable(index, WARNING);
+	idsTable = PGrnLookupIDsTable(index, WARNING);
 
 	if (!stats)
-		stats = GrnBulkDeleteResult(info, idsTable);
+		stats = PGrnBulkDeleteResult(info, idsTable);
 
 	if (!idsTable || !callback)
 		PG_RETURN_POINTER(stats);
@@ -1127,9 +1496,7 @@ pgroonga_bulkdelete(PG_FUNCTION_ARGS)
 			ctid = UInt64ToCtid(*key);
 			if (callback(&ctid, callback_state))
 			{
-				GrnLock(index, ExclusiveLock);
 				grn_table_cursor_delete(ctx, cursor);
-				GrnUnlock(index, ExclusiveLock);
 
 				nRemovedTuples += 1;
 			}
@@ -1149,10 +1516,10 @@ pgroonga_bulkdelete(PG_FUNCTION_ARGS)
 }
 
 static void
-GrnRemoveUnusedTables(void)
+PGrnRemoveUnusedTables(void)
 {
 	grn_table_cursor *cursor;
-	const char *min = GrnIDsTableNamePrefix;
+	const char *min = PGrnIDsTableNamePrefix;
 
 	cursor = grn_table_cursor_open(ctx, grn_ctx_db(ctx),
 								   min, strlen(min),
@@ -1180,7 +1547,7 @@ GrnRemoveUnusedTables(void)
 			grn_obj *table;
 
 			snprintf(tableName, sizeof(tableName),
-					 GrnLexiconNameFormat, relationID);
+					 PGrnLexiconNameFormat, relationID);
 			table = grn_ctx_get(ctx, tableName, strlen(tableName));
 			if (table)
 			{
@@ -1188,7 +1555,7 @@ GrnRemoveUnusedTables(void)
 			}
 
 			snprintf(tableName, sizeof(tableName),
-					 GrnIDsTableNameFormat, relationID);
+					 PGrnIDsTableNameFormat, relationID);
 			table = grn_ctx_get(ctx, tableName, strlen(tableName));
 			if (table)
 			{
@@ -1210,10 +1577,10 @@ pgroonga_vacuumcleanup(PG_FUNCTION_ARGS)
 	IndexBulkDeleteResult *stats = (IndexBulkDeleteResult *) PG_GETARG_POINTER(1);
 
 	if (!stats)
-		stats = GrnBulkDeleteResult(info,
-									GrnLookupIDsTable(info->index, WARNING));
+		stats = PGrnBulkDeleteResult(info,
+									 PGrnLookupIDsTable(info->index, WARNING));
 
-	GrnRemoveUnusedTables();
+	PGrnRemoveUnusedTables();
 
 	PG_RETURN_POINTER(stats);
 }
@@ -1237,5 +1604,24 @@ pgroonga_costestimate(PG_FUNCTION_ARGS)
 Datum
 pgroonga_options(PG_FUNCTION_ARGS)
 {
-	return (Datum) 0;
+	Datum reloptions = PG_GETARG_DATUM(0);
+	bool validate = PG_GETARG_BOOL(1);
+	relopt_value *options;
+	PGrnOptions *grnOptions;
+	int nOptions;
+	const relopt_parse_elt optionsMap[] = {
+		{"tokenizer", RELOPT_TYPE_STRING,
+		 offsetof(PGrnOptions, tokenizerOffset)},
+		{"normalizer", RELOPT_TYPE_STRING,
+		 offsetof(PGrnOptions, normalizerOffset)}
+	};
+
+	options = parseRelOptions(reloptions, validate, PGrnReloptionKind,
+							  &nOptions);
+	grnOptions = allocateReloptStruct(sizeof(PGrnOptions), options, nOptions);
+	fillRelOptions(grnOptions, sizeof(PGrnOptions), options, nOptions,
+				   validate, optionsMap, lengthof(optionsMap));
+	pfree(options);
+
+	PG_RETURN_BYTEA_P(grnOptions);
 }
